"""
OWLS definitions




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

ghdf5_ptypes  = ("PartType0", "PartType1", "PartType2", "PartType3",
                 "PartType4", "PartType5")
